#include "defs.h"

/*
 * from /usr/include/linux/netlink.h
 */
#define NLMSG_ALIGNTO 4
#define NLMSG_DATA(nlh) (nlh + roundup(STRUCT_SIZE("nlmsghdr"), NLMSG_ALIGNTO))

extern char *help_dumpaudit[];
extern void cmd_dumpaudit(void);

static void dump_audit_skb_queue(ulong audit_skb_queue);
static void __do_dumpaudit(char *symname);
static void do_dumpaudit(void);

static struct command_table_entry command_table[] = {
	{ "dumpaudit", cmd_dumpaudit, help_dumpaudit, 0 },
	{ (char *)NULL }
};

static int flag_debug;

int
_init(void)
{
	register_extension(command_table);
	return 1;
}

int
_fini(void)
{
	return 1;
}

char *help_dumpaudit[] = {
"dumpaudit [-d]",
"dumpaudit - dump audit events left in the kernel audit buffer",
"  ",
"  dumpaudit",
"  ",
"  This command retrieves audit events left in the kernel audit buffer and dumps them.",
"  ",
"       -d print debug information",
NULL,
};

void
cmd_dumpaudit(void)
{
	int c;

	flag_debug = FALSE;

	while ((c = getopt(argcnt, args, "d")) != EOF) {
		switch (c) {
		case 'd':
			flag_debug = TRUE;
			break;
		default:
			argerrs++;
			break;
		}
	}

	if (argerrs)
		cmd_usage(pc->curcmd, SYNOPSIS);

	do_dumpaudit();
}

static void dump_audit_skb_queue(ulong audit_skb_queue)
{
	ulong skb_buff_head_next, p;
	uint32_t qlen;

	readmem(audit_skb_queue + MEMBER_OFFSET("sk_buff_head", "qlen"),
		KVADDR,
		&qlen,
		sizeof(qlen),
		"audit_skb_queue.qlen",
		FAULT_ON_ERROR);

	if (!qlen)
		return;

	readmem(audit_skb_queue,
		KVADDR,
		&skb_buff_head_next,
		sizeof(skb_buff_head_next),
		"audit_skb_queue.next",
		FAULT_ON_ERROR);

	if (!skb_buff_head_next)
		return;

	p = skb_buff_head_next;
	do {
		ulong data;

		if (flag_debug)
			fprintf(fp, "%#016lx\n", p);

		readmem(p + MEMBER_OFFSET("sk_buff", "data"),
			KVADDR,
			&data,
			sizeof(data),
			"sk_buff.data",
			FAULT_ON_ERROR);

		if (data) {
			uint32_t nlmsg_len;
			uint16_t nlmsg_type;
			char *buf = NULL;

			readmem(data + MEMBER_OFFSET("nlmsghdr", "nlmsg_len"),
				KVADDR,
				&nlmsg_len,
				MEMBER_SIZE("nlmsghdr", "nlmsg_len"),
				"nlmsghdr.nlmsg_len",
				FAULT_ON_ERROR);
			readmem(data + MEMBER_OFFSET("nlmsghdr", "nlmsg_type"),
				KVADDR,
				&nlmsg_type,
				MEMBER_SIZE("nlmsghdr", "nlmsg_type"),
				"nlmsghdr.nlmsg_type",
				FAULT_ON_ERROR);

			buf = GETBUF(nlmsg_len + 1);
			readmem(NLMSG_DATA(data),
				KVADDR,
				buf,
				nlmsg_len,
				"sk_buff.data + sizeof(struct nlmsghdr)",
				FAULT_ON_ERROR);
			buf[nlmsg_len] = '\0';

			fprintf(fp, "type=%u %s\n", nlmsg_type, buf);
			FREEBUF(buf);
		}

		readmem(p,
			KVADDR,
			&p,
			sizeof(p),
			"skb_buff.next",
			FAULT_ON_ERROR);
	} while (p != audit_skb_queue);
}

static void __do_dumpaudit(char *symname)
{
	if (symbol_exists(symname)) {
		if (flag_debug)
			fprintf(fp, "# %s:\n", symname);
		dump_audit_skb_queue(symbol_value(symname));
	}
}

static void do_dumpaudit(void)
{
	__do_dumpaudit("audit_skb_hold_queue");
	__do_dumpaudit("audit_skb_queue");
}
